from enum import Enum, unique


@unique
class MessageNotificationLevel(Enum):
    ALL_MESSAGES = 0
    ONLY_MENTIONS = 1


@unique
class ExplicitContentFilterLevel(Enum):
    DISABLED = 0
    MEMBERS_WITHOUT_ROLES = 1
    ALL_MEMBERS = 2


@unique
class MFALevel(Enum):
    NONE = 0
    ELEVATED = 1


@unique
class VerificationLevel(Enum):
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    VERY_HIGH = 4


@unique
class ChannelTypes(Enum):
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4


@unique
class MessageActivityTypes(Enum):
    JOIN = 1
    SPECTATE = 2
    LISTEN = 3
    JOIN_REQUEST = 5


@unique
class GatewayOpcodes(Enum):
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    STATUS_UPDATE = 3
    VOICE_STATE_UPDATE = 4
    VOICE_SERVER_PING = 5
    RESUME = 6
    RECONNECT = 7
    REQUEST_GUILD_MEMBERS = 8
    INVALID_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11



import sys

class WebSocketCreationError(Exception):
    def __init__(self):
        print("Error creating a websocket to connect to discord.", file=sys.stderr)

class EventTypeError(Exception):
    pass

class AuthorizationError(Exception):
    def __init__(self):
        print('You requested a api endpoint which you have no authorization', file=sys.stderr)

class UnhandledEndpointStatusError(Exception):
    pass


import aiohttp
import asyncio
import json
import time
import signal
from threading import Thread

from .exceptions import EventTypeError
from .user import User, UserConnection
from .guild import Guild, GuildMember
from .base import DiscordObject
from .constants import DISCORD_API_URL
from .channel import Channel
from .http import HTTPHandler
from .websocket import DiscordWebsocket

import logging
logger = logging.getLogger(__name__)


def task_handler():
    logger.debug('Stopping tasks')
    for task in asyncio.Task.all_tasks():
        task.cancel()


class DiscordBot:
    def __init__(self, token):
        """DiscordBot constructor.

        Args:
            token (str): The bot discord token.
        """
        self.token: str = token
        self.http: HTTPHandler = HTTPHandler(token, self)
        self.guilds = []
        self.loop = asyncio.get_event_loop()
        self.do_sync = self.loop.run_until_complete
        self.loop.add_signal_handler(signal.SIGINT, task_handler)
        self.user: User = User()
        self.ws: DiscordWebsocket = None

    def run(self):
        try:
            self.do_sync(self.start())
        except KeyboardInterrupt:
            # TODO: Add bot close code here
            pass
        except asyncio.CancelledError:
            logger.debug('Tasks has been cancelled')
        finally:
            self.do_sync(self.http.close_session())
            self.loop.close()

    async def raise_event(self, event: str, *args, **kwargs):
        try:
            if hasattr(self, event):
                await getattr(self, event)(*args, **kwargs)
        except asyncio.CancelledError as e:
            logger.error(e, exc_info=1)
            pass
        except Exception as e:
            logger.error(e, exc_info=1)
            # TODO: handle error here
            pass

    def event(self, name: str=None):
        """DiscordBot event decorator, uses the function's name or the 'name' parameter to subscribe to a event"""
        def real_event(coro):
            if not asyncio.iscoroutinefunction(coro):
                raise EventTypeError(
                    'The event function must be a coroutine.')
            if name is not None:
                if not isinstance(name, str):
                    raise TypeError('event name must be of type str')
                else:
                    try:
                        getattr(self, name)
                        raise AttributeError(
                            'tried to subscribe to a event that doesn\'t exist, or you already subscribed to it.')
                    except AttributeError:
                        pass
                    finally:
                        setattr(self, name, coro)
            else:
                setattr(self, coro.__name__, coro)
            logger.debug(f'{coro.__name__} subscribed succesfully.')
            return coro
        return real_event

    async def start(self):
        await self.http.create_session()
        self.ws = DiscordWebsocket(self.http)
        await self.ws.start()
    
    async def exit(self):
        if not self.ws.closed:
            await self.ws.close()
        await self.http.close_session()
        

    async def change_avatar(self, url: str):
        raise NotImplementedError()
        # TODO: Implement this: https://discordapp.com/developers/docs/resources/user#modify-current-user

    async def get_user(self, id: int) -> User:
        res = await self.http.request_url('/users/' + str(id))
        return await User.from_api_res(res)

    async def get_self_user(self) -> User:
        return await self.get_user("@me")

    async def get_guilds(self) -> list:
        res = await self.http.request_url('/users/@me/guilds')
        return await Guild.from_api_res(res)

    async def get_guild(self, id) -> Guild:
        res = await self.http.request_url('/guilds/' + str(id))
        return await Guild.from_api_res(res)

    async def get_guild_members(self, guild: Guild):
        """Gets and fills the guild with it's members info"""

        res = await self.http.request_url('/guilds/' + guild.id + '/members')
        await guild._fill_members(res)

    async def get_guild_member(self, guild: Guild, member_id: int):
        res = await self.http.request_url('/guilds/' + guild.id + '/members/' + member_id)
        return await GuildMember.from_api_res(res)

    async def leave_guild(self, guild_id: int):
        await self.http.request_url(f'/users/@me/guilds/{guild_id}', type='DELETE')

    async def get_channel(self, channel_id: int):
        res = await self.http.request_url(f'/channels/{channel_id}')
        return await Channel.from_api_res(res)

    async def delete_channel(self, channel_id: int):
        """Delete a channel, or close a private message. Requires the 'MANAGE_CHANNELS' permission for the guild. 
        Deleting a category does not delete its child channels; they will have their parent_id removed and a Channel Update Gateway event will fire for each of them. 
        Returns a channel object on success. 
        Fires a Channel Delete Gateway event."""
        res = await self.http.request_url(f'/channels/{channel_id}', type='DELETE')
        return await Channel.from_api_res(res)

    async def get_dms(self) -> list:
        res = await self.http.request_url('/users/@me/channels')
        return await Channel.from_api_res(res)

    async def get_connections(self) -> list:
        res = await self.http.request_url('/users/@me/connections')
        return await UserConnection.from_api_res(res)



import json

from .base import DiscordObject
from .user import User
from .emoji import Emoji
from .constants import DISCORD_CDN
from .channel import Channel
from .role import Role

import logging
logger = logging.getLogger(__name__)


class GuildEmbed(DiscordObject):
    """A Guild Embed Object"""

    def __init__(self, enabled=False, channel_id=0):
        self.enabled = enabled
        self.channel_id = channel_id


class GuildMember(DiscordObject):
    def __init__(self, user=User(), nick="", roles=[], joined_at=None, deaf=False,
                 mute=False):
        self.user = user
        self.nick = nick
        self.roles = roles
        self.joined_at = joined_at
        self.deaf = deaf
        self.mute = mute

    async def _from_api_ext(self, key, value):
        if key == 'user':
            setattr(self, key, await User.from_api_res(value))
        else:
            await super()._from_api_ext(key, value)

    def __str__(self):
        return self.user.__str__()

    def __repr__(self):
        return f'<GuildMember Object: {self.user.username}#{self.user.discriminator}>'


class Guild(DiscordObject):
    """A Discord Guild Object"""

    def __init__(self, id=0, name="", icon="", splash="", owner=False,
                 owner_id=0, permissions=0, region="", afk_channel_id=0,
                 afk_timeout=0, embed_enabled=False, embed_channel_id=0, verification_level=0,
                 default_message_notifications=0, explicit_content_filter=0, roles=[], emojis=[],
                 features=[], mfa_level=0, application_id=0, widget_enabled=False,
                 widget_channel_id=0, system_channel_id=0, joined_at=None, large=None,
                 unavailable=None, member_count=None, voice_states=None, members=[],
                 channels=None, presences=None):
        self.id = id
        self.name = name
        self.icon = icon
        self.splash = splash
        self.owner = owner
        self.owner_id = owner_id
        self.permissions = permissions
        self.region = region
        self.afk_channel_id = afk_channel_id
        self.afk_timeout = afk_timeout
        self.embed_enabled = embed_enabled
        self.embed_channel_id = embed_channel_id
        self.verification_level = verification_level
        self.default_message_notifications = default_message_notifications
        self.explicit_content_filter = explicit_content_filter
        self.roles = roles
        self.emojis = emojis
        self.features = features
        self.mfa_level = mfa_level
        self.application_id = application_id
        self.widget_enabled = widget_enabled
        self.widget_channel_id = widget_channel_id
        self.system_channel_id = system_channel_id
        self.joined_at = joined_at
        self.large = large
        self.unavailable = unavailable
        self.member_count = member_count
        self.voice_states = voice_states
        self.members = members
        self.channels = channels
        self.presences = presences

    async def _from_api_ext(self, key, value):
        if key == 'roles':
            setattr(self, key, [await Role.from_api_res(x) for x in value])
            pass
        elif key == 'members':
            setattr(self, key, [await GuildMember.from_api_res(x) for x in value])
            pass
        elif key == 'emojis':
            setattr(self, key, [await Emoji.from_api_res(x) for x in value])
            pass
        elif key == 'channels':
            setattr(self, key, [await Channel.from_api_res(x) for x in value])
            pass
        else:
            await super()._from_api_ext(key, value)

    async def _fill_members(self, members: list):
        self.members = []
        for member in members:
            self.members.append(await GuildMember.from_api_res(member))

    def is_owner(self, member: GuildMember):
        return self.owner_id == member.user.id

    def get_icon(self):
        return DISCORD_CDN + f'/icons/{self.id}/{self.icon}.png'

    def get_splash(self):
        return DISCORD_CDN + f'/splashes/{self.id}/{self.splash}.png'

    def __str__(self):
        return f'{self.name}#{self.id}'

    def __repr__(self):
        return f'<Guild Object: {self.name}#{self.id}>'


class IntegrationAccount(DiscordObject):
    def __init__(self, id="", name=""):
        self.id = id
        self.name = name


class Integration(DiscordObject):
    def __init__(self, id=0, name="", type="", enabled=False, syncing=False,
                 role_id=0, expire_behavior=0, expire_grace_period=0, user=None,
                 account=None, synced_at=None):
        self.id = id
        self.name = name
        self.type = type
        self.enabled = enabled
        self.syncing = syncing
        self.role_id = role_id
        self.expire_behavior = expire_behavior
        self.expire_grace_period = expire_grace_period
        self.user = user
        self.account = account
        self.synced_at = synced_at

    async def _from_api_ext(self, key, value):
        if key == 'user':
            setattr(self, key, await User.from_api_res(value))
        elif key == 'account':
            setattr(self, key, await IntegrationAccount.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)


class Ban(DiscordObject):
    def __init__(self, reason="", user=None):
        self.reason = reason
        self.user = user
    
    async def _from_api_ext(self, key, value):
        if key == 'user':
            setattr(self, key, await User.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)



"""Users in Discord are generally considered the base entity. Users can spawn across the entire platform,
be members of guilds, participate in text and voice chat, and much more. Users are separated by a distinction of "bot" vs "normal." 
Although they are similar, bot users are automated users that are "owned" by another user. Unlike normal users, 
bot users do not have a limitation on the number of Guilds they can be a part of."""

from .base import DiscordObject
from .constants import DISCORD_CDN

class UserConnection(DiscordObject):
    """A Discord User Connection"""

    def __init__(self, id="", name="", type="", revoked=False, integrations=[]):
        self.id = id
        self.name = name
        self.type = type
        self.revoked = revoked
        self.integrations = integrations


class User(DiscordObject):
    """A Discord User Object"""

    def __init__(self, id=0, username="", discriminator="", avatar="", bot=False,
                 mfa_enabled=False, verified=False, email=""):
        self.id = id
        self.username = username
        self.discriminator = discriminator
        self.avatar = avatar
        self.bot = bot
        self.mfa_enabled = mfa_enabled
        self.verified = verified
        self.email = email

    def __str__(self):
        return f'{self.username}#{self.discriminator}'

    def __repr__(self):
        return f'<User Object: {self.username}#{self.discriminator}, {self.id}>'
    
    def get_avatar_url(self):
        return DISCORD_CDN + f'/avatars/{self.id}/{self.avatar}.png'

    def get_default_avatar_url(self):
        return DISCORD_CDN + f'/embed/avatars/{int(self.discriminator) % 5}.png'
    
    async def _from_api_ext(self, key, value):
        if key == 'guild_id':
            pass
        else:
            await super()._from_api_ext(key, value)


"""Contains all related Channel Discord objects"""

from .base import DiscordObject
from .user import User
from .role import Role
from .emoji import Emoji


class Overwrite(DiscordObject):
    def __init__(self, id=0, type="", allow=0, deny=0):
        self.id = id
        self.type = type
        self.allow = allow
        self.deny = deny


class Channel(DiscordObject):
    """Represents a guild or DM channel within Discord."""

    def __init__(self, id=0, type=0, guild_id=0, position=0, permission_overwrites=[],
                 name="", topic="", nsfw=False, last_message_id=0,
                 bitrate=0, user_limit=0, recipients=[], icon="",
                 owner_id=0, application_id=0, parent_id=0, last_pin_timestamp=None):
        self.id = id
        self.type = type
        self.guild_id = guild_id
        self.position = position
        self.permission_overwrites = permission_overwrites
        self.name = name
        self.topic = topic
        self.nsfw = nsfw
        self.last_message_id = last_message_id
        self.bitrate = bitrate
        self.user_limit = user_limit
        self.recipients = recipients
        self.icon = icon
        self.owner_id = owner_id
        self.application_id = application_id
        self.parent_id = parent_id
        self.last_pin_timestamp = last_pin_timestamp

    async def _from_api_ext(self, key, value):
        if key == 'recipients':
            setattr(self, key, [await User.from_api_res(x) for x in value])
        elif key == 'permission_overwrites':
            setattr(self, key, [await Overwrite.from_api_res(x) for x in value])
        else:
            return await super()._from_api_ext(key, value)
    
    def __str__(self):
        return f'{self.name}#{self.id}'

    def __repr__(self):
        return f'<Channel Object: {self.name}#{self.id}>'


class MessageActivity(DiscordObject):
    def __init__(self, type=None, party_id=""):
        self.type = type
        self.party_id = party_id


class MessageApplication(DiscordObject):
    def __init__(self, id=0, cover_image="", description="", icon="", name=""):
        self.id = id
        self.cover_image = cover_image
        self.description = description
        self.icon = icon
        self.name = name


class Reaction(DiscordObject):
    def __init__(self, count=0, me=False, emoji=None):
        self.count = count
        self.me = me
        self.emoji = emoji

    async def _from_api_ext(self, key, value):
        if key == 'emoji':
            setattr(self, key, await Emoji.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)


class EmbedThumbnail(DiscordObject):
    def __init__(self, url="", proxy_url="", height=0, width=0):
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width


class EmbedVideo(DiscordObject):
    def __init__(self, url="", height=0, width=0):
        self.url = url
        self.height = height
        self.width = width


class EmbedImage(DiscordObject):
    def __init__(self, url="", proxy_url="", height=0, width=0):
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width


class EmbedProvider(DiscordObject):
    def __init__(self, name="", url=""):
        self.name = name
        self.url = url


class EmbedAuthor(DiscordObject):
    def __init__(self, name="", url="", icon_url="", proxy_icon_url=""):
        self.name = name
        self.url = url
        self.icon_url = icon_url
        self.proxy_icon_url = proxy_icon_url


class EmbedFooter(DiscordObject):
    def __init__(self, text="", icon_url="", proxy_icon_url=""):
        self.text = text
        self.icon_url = icon_url
        self.proxy_icon_url = proxy_icon_url


class EmbedField(DiscordObject):
    def __init__(self, name="", value="", inline=False):
        self.name = name
        self.value = value
        self.inline = inline


class Embed(DiscordObject):
    def __init__(self, title="", type="", description="", url="", timestamp=None,
                 color=0, footer=EmbedFooter(), image=EmbedImage(), thumbnail=EmbedThumbnail(),
                 video=EmbedVideo(), provider=EmbedProvider(), author=EmbedAuthor(), fields=[]):
        self.title = title
        self.type = type
        self.description = description
        self.url = url
        self.timestamp = timestamp
        self.color = color
        self.footer = footer
        self.image = image
        self.thumbnail = thumbnail
        self.video = video
        self.provider = provider
        self.author = author
        self.fields = fields

    async def _from_api_ext(self, key, value):
        if key == 'footer':
            setattr(self, key, await EmbedFooter.from_api_res(value))
        elif key == 'image':
            setattr(self, key, await EmbedImage.from_api_res(value))
        elif key == 'thumbnail':
            setattr(self, key, await EmbedThumbnail.from_api_res(value))
        elif key == 'video':
            setattr(self, key, await EmbedVideo.from_api_res(value))
        elif key == 'provider':
            setattr(self, key, await EmbedProvider.from_api_res(value))
        elif key == 'author':
            setattr(self, key, await EmbedAuthor.from_api_res(value))
        elif key == 'fields':
            setattr(self, key, [await EmbedField.from_api_res(x) for x in value])
        else:
            return await super()._from_api_ext(key, value)


class Attachment(DiscordObject):
    def __init__(self, id=0, filename="", size=0, url="", proxy_url="",
                 height=0, width=0):
        self.id = id
        self.filename = filename
        self.size = size
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width


class ChannelMessage(DiscordObject):
    """Represents a message sent in a channel within Discord.

    The author object follows the structure of the user object, but is only a valid user in the case where the message is generated by a user or bot user. 
    If the message is generated by a webhook, the author object corresponds to the webhook's id, username, and avatar. 
    You can tell if a message is generated by a webhook by checking for the webhook_id on the message object."""

    def __init__(self, id=0, channel_id=0, author=None, content="", timestamp=None,
                 edited_timestamp=None, tts=False, mention_everyone=False, mentions=[],
                 mention_roles=[], attachments=[], embeds=[], reactions=[],
                 nonce=0, pinned=False, webhook_id=0, type=0,
                 activity=MessageActivity(), application=MessageApplication()):
        self.id = id
        self.channel_id = channel_id
        self.author = author
        self.content = content
        self.timestamp = timestamp
        self.edited_timestamp = edited_timestamp
        self.tts = tts
        self.mention_everyone = mention_everyone
        self.mentions = mentions
        self.mention_roles = mention_roles
        self.attachments = attachments
        self.embeds = embeds
        self.reactions = reactions
        self.nonce = nonce
        self.pinned = pinned
        self.webhook_id = webhook_id
        self.type = type
        self.activity = activity
        self.application = application

    async def _from_api_ext(self, key, value):
        if key == 'author':
            setattr(self, key, await User.from_api_res(value))
        elif key == 'mentions':
            setattr(self, key, [await User.from_api_res(x) for x in value])
        elif key == 'mention_roles':
            setattr(self, key, [await Role.from_api_res(x) for x in value])
        elif key == 'attachments':
            setattr(self, key, [await Attachment.from_api_res(x) for x in value])
        elif key == 'reactions':
            setattr(self, key, [await Reaction.from_api_res(x) for x in value])
        elif key == 'activity':
            setattr(self, key, await MessageActivity.from_api_res(value))
        elif key == 'activity':
            setattr(self, key, await MessageApplication.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)



import json
import asyncio
import aiohttp

import logging
logger = logging.getLogger(__name__)

class DiscordObject:
    """Base class for discord objects."""

    @classmethod
    async def from_api_res(cls, coro_or_json_or_str):
        """Parses a discord API response"""
        # TODO: Maybe make this asynchronous, search for async_generator object and how to handle them

        json_obj = coro_or_json_or_str

        if isinstance(coro_or_json_or_str, str):
            json_obj = json.loads(coro_or_json_or_str)
        elif asyncio.iscoroutine(coro_or_json_or_str):
            json_obj = await coro_or_json_or_str()
        elif isinstance(coro_or_json_or_str, aiohttp.ClientResponse):
            json_obj = await coro_or_json_or_str.json()

        if isinstance(json_obj, list):
            lst = []
            for item in json_obj:
                result = cls()
                for key, value in item.items():
                    if hasattr(result, key):
                        await result._from_api_ext(key, value)
                lst.append(result)
            return lst
        elif isinstance(json_obj, dict):
            result = cls()
            for key, value in json_obj.items():
                if hasattr(result, key):
                    await result._from_api_ext(key, value)
            return result
        else:
            raise ValueError('it must be a dictionary or a list.')

    async def _from_api_ext(self, key, value):
        """Api response decoding extensions, called automatically by DiscordObject.from_api_res().
        Used if the class contains a attribute that it's a class and must be initialized with info,
        also used if the class contains an array of classes as attribute."""
        setattr(self, key, value)



from .base import DiscordObject
from .guild import Guild
from .channel import Channel
from .user import User

class Invite(DiscordObject):
    """Represents a code that when used, adds a user to a guild."""

    def __init__(self, code="", guild=None, channel=None):
        self.code = code
        self.guild = guild
        self.channel = channel
    
    async def _from_api_ext(self, key, value):
        if key == 'guild':
            setattr(self, key, await Guild.from_api_res(value))
        elif key == 'channel':
            setattr(self, key, await Channel.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)


class InviteMetadata(DiscordObject):
    def __init__(self, inviter=None, uses=0, max_uses=0, max_age=0, temporary=False,
                 created_at=None, revoked=False):
        self.inviter = inviter
        self.uses = uses
        self.max_uses = max_uses
        self.max_age = max_age
        self.temporary = temporary
        self.created_at = created_at
        self.revoked = revoked

    async def _from_api_ext(self, key, value):
        if key == 'inviter':
            setattr(self, key, await User.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)


import aiohttp
import asyncio
import json
import logging
import platform

from .user import User, UserConnection
from .guild import Guild, GuildMember
from .base import DiscordObject
from .version import VERSION_STR
from .constants import DISCORD_API_URL
from .channel import Channel, ChannelMessage
from .emoji import Emoji
from .exceptions import WebSocketCreationError, AuthorizationError, UnhandledEndpointStatusError
from .enums import GatewayOpcodes

import logging
logger = logging.getLogger(__name__)


class RateLimit(DiscordObject):
    def __init__(self, message="", retry_after=0, _global=False):
        self.message = message
        self.retry_after = retry_after
        self._global = _global


class HTTPHandler:
    def __init__(self, token, discord_client):
        self.token: str = token
        self.loop = asyncio.get_event_loop()
        self.headers: dict = None
        self.update_headers()
        self.session: aiohttp.ClientResponse = None
        self.discord_client = discord_client

    async def create_session(self):
        self.update_headers()
        self.session = aiohttp.ClientSession(
            headers=self.headers, auto_decompress=True)

    async def close_session(self):
        await self.session.close()
        logger.debug('Session closed!')

    def update_headers(self):
        self.headers = {'Authorization': 'Bot ' + self.token,
                        'User-Agent': f'DiscordBot (https://github.com/Ryozuki/discord.aio, {VERSION_STR})'}
    
    def get_client(self):
        return self.discord_client

    async def request_url(self, url, type='GET', data=None):
        while True:
            operation = None
            if type == 'GET':
                operation = self.session.get(DISCORD_API_URL + url)
            elif type == 'POST':
                operation = self.session.post(DISCORD_API_URL + url, data=data)
            elif type == 'DELETE':
                operation = self.session.delete(DISCORD_API_URL + url)

            async with operation as res:
                if res.status == 429:
                    limit = await RateLimit.from_api_res(res)

                    if 'X-RateLimit-Remaining' in res.headers and int(res.headers['X-RateLimit-Remaining']) > 0:
                        logger.debug(
                            f"Status is {res.status} but i have {res.headers['X-RateLimit-Remaining']} ratelimits remaining, ")
                        continue

                    logger.debug(
                        f"Status is {res.status} so we must wait {limit.retry_after / 1000} seconds!")
                    await asyncio.sleep(limit.retry_after / 1000)
                    logger.debug("Done waiting! Requesting again")
                elif res.status < 300 and res.status >= 200:
                    return await res.json()
                elif res.status == 401:
                    raise AuthorizationError
                else:
                    raise UnhandledEndpointStatusError



"""Webhooks are a low-effort way to post messages to channels in Discord. They do not require a bot user or authentication to use."""

from .base import DiscordObject


class Webhook(DiscordObject):
    """Used to represent a webhook."""

    def __init__(self, id=0, guild_id=0, channel_id=0, user=None, name="",
                 avatar="", token=""):
        self.id = id
        self.guild_id = guild_id
        self.channel_id = channel_id
        self.user = user
        self.name = name
        self.avatar = avatar
        self.token = token



"""
discord.aio is an asynchronous Discord API wrapper for python 3.6+
"""

from .version import VERSION_STR

__author__ = 'Ryozuki'
__license__ = 'MIT'
__version__ = VERSION_STR

from .client import DiscordBot
from .channel import Channel, ChannelMessage, Attachment, Embed, EmbedAuthor, EmbedField, EmbedFooter, EmbedImage, EmbedProvider, EmbedThumbnail, EmbedVideo
from .emoji import Emoji
from .user import User, UserConnection
from .guild import Guild, GuildEmbed, GuildMember, Integration, IntegrationAccount
from .role import Role
from .http import HTTPHandler
from .websocket import DiscordWebsocket
from .webhook import Webhook
from .invite import Invite, InviteMetadata
from .enums import ChannelTypes, ExplicitContentFilterLevel, MessageActivityTypes, MessageNotificationLevel, MFALevel, VerificationLevel
from .exceptions import WebSocketCreationError, AuthorizationError, EventTypeError, UnhandledEndpointStatusError
from .base import DiscordObject
from .voice import VoiceRegion, VoiceState


from .base import DiscordObject


class VoiceState(DiscordObject):
    def __init__(self, guild_id=0, channel_id=0, user_id=0, session_id="", deaf=False,
                 mute=False, self_deaf=False, self_mute=False, suppress=False):
        self.guild_id = guild_id
        self.channel_id = channel_id
        self.user_id = user_id
        self.session_id = session_id
        self.deaf = deaf
        self.mute = mute
        self.self_deaf = self_deaf
        self.self_mute = self_mute
        self.suppress = suppress


class VoiceRegion(DiscordObject):
    def __init__(self, id="", name="", vip=False, optimal=False, deprecated=False,
                 custom=False):
        self.id = id
        self.name = name
        self.vip = vip
        self.optimal = optimal
        self.deprecated = deprecated
        self.custom = custom



from .base import DiscordObject


class Role(DiscordObject):
    def __init__(self, id=0, name="", color=0, hoist=False, position=0,
                 permissions=0, managed=False, mentionable=False):
        self.id = id
        self.name = name
        self.color = color
        self.hoist = hoist
        self.position = position
        self.permissions = permissions
        self.managed = managed
        self.mentionable = mentionable

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'<Role Object: {self.name}#{self.id}>'



import json

from .base import DiscordObject
from .user import User
from .constants import DISCORD_CDN


class Emoji(DiscordObject):
    def __init__(self, id=0, name="", roles=[], user=None, require_colons=False,
                 managed=False, animated=False):
        self.id = id
        self.name = name
        self.roles = roles
        self.user = user
        self.require_colons = require_colons
        self.managed = managed
        self.animated = animated

    async def _from_api_ext(self, key, value):
        # Note: roles here is only an array of ids
        if key == 'user':
            setattr(self, key, await User.from_api_res(value))
        else:
            return await super()._from_api_ext(key, value)

    def get_url(self):
        return DISCORD_CDN + f'/emojis/{self.id}.png'



import asyncio
import aiohttp
import json
import platform

from .enums import GatewayOpcodes
from .guild import Guild
from .user import User
from .http import HTTPHandler
from .channel import Channel, ChannelMessage
from .emoji import Emoji

import logging
logger = logging.getLogger(__name__)


class DiscordWebsocket:
    def __init__(self, http: HTTPHandler=None, session_id: str=None, shards: list=[]):
        self.heartbeat_interval: float = None
        self._trace: list = []
        self.seq: str = None
        self.heartbeat_future: asyncio.Future = None
        self.session_id: str = session_id
        self.http: HTTPHandler = http
        self.gateway_url: str = None
        self.shards: list = shards
        self.ws: aiohttp.ClientWebSocketResponse = None

    @property
    def closed(self) -> bool:
        return self.ws is None or self.ws.closed

    async def send_heartbeat(self):
        while True:
            if self.ws is None or self.ws.closed:
                break
            logger.debug(
                f'Waiting {self.heartbeat_interval / 1000} seconds to send heartbeat')
            await asyncio.sleep(self.heartbeat_interval / 1000)
            await self.ws.send_json({
                'op': 1,
                'd': self.s
            })
            logger.debug("Sent heartbeat")

    async def close(self) -> bool:
        if self.ws is not None and not self.ws.closed:
            if self.heartbeat_future is not None and not self.heartbeat_future.cancelled():
                self.heartbeat_future.cancel()
            await self.ws.close()
            logger.debug('Websocket closed!')
            return True
        else:
            return False

    async def start(self):
        if self.ws is not None and not self.ws.closed:
            await self.ws.close()
            self.ws = None

        info = await self.http.request_url('/gateway/bot')
        self.gateway_url = info['url']
        self.shards = info['shards']

        logger.debug(f'I can use {self.shards} shards!')

        async with self.http.session.ws_connect(self.gateway_url + '?v=6&encoding=json') as ws:
            self.ws = ws
            async for msg in self.ws:
                # logger.debug(msg)
                if msg.type == aiohttp.WSMsgType.TEXT:
                    dct = json.loads(msg.data)
                    opcode = dct['op']
                    data = dct['d']
                    if GatewayOpcodes(opcode) == GatewayOpcodes.HELLO:
                        if self.heartbeat_future is not None:
                            if not self.heartbeat_future.cancelled():
                                self.heartbeat_future.cancel()
                                self.heartbeat_future = None
                        self.heartbeat_interval = dct['d']['heartbeat_interval']
                        self._trace = dct['d']['_trace']
                        self.seq = dct['s']
                        if self.session_id is not None:
                            pass
                            await ws.send_json({
                                "op": 6,  # Resume
                                "d": {
                                    "token": self.http.token,
                                    "session_id": self.session_id,
                                    "seq": self.seq
                                }
                            })
                        else:
                            await ws.send_json({
                                "op": 2,  # Identify
                                "d": {
                                    "token": self.http.token,
                                    "properties": {
                                        '$os': platform.system(),
                                        '$browser': 'discord.aio',
                                        '$device': 'discord.aio'
                                    },
                                    "compress": False,
                                    "large_threshold": 250
                                }
                            })
                        logger.debug(
                            f'Ensuring to heartbeat every {self.heartbeat_interval / 1000} seconds!')
                        self.heartbeat_future = asyncio.ensure_future(
                            self.send_heartbeat())
                    elif GatewayOpcodes(opcode) == GatewayOpcodes.HEARTBEAT_ACK:
                        logger.debug(
                            f'Got HEARTBEAT_ACK (new s: {dct["s"]}, old s: {self.seq})')
                        self.s = dct['s']
                    elif GatewayOpcodes(opcode) == GatewayOpcodes.DISPATCH:
                        event_type = dct['t']
                        asyncio.ensure_future(
                            self.dispatch_event(event_type, data))

    async def dispatch_event(self, event, data):
        if event == 'READY':
            self.session_id = data['session_id']
            self.http.get_client().user = await User.from_api_res(data['user'])
            self.http.get_client().guilds = await Guild.from_api_res(data['guilds'])
            asyncio.ensure_future(
                self.http.get_client().raise_event('on_ready'))

        elif event == 'GUILD_CREATE':
            guild = await Guild.from_api_res(data)
            return await self.http.get_client().raise_event('on_guild_create', guild)

        elif event == 'GUILD_DELETE':
            guild = await Guild.from_api_res(data)
            return await self.http.get_client().raise_event('on_guild_delete', guild)

        elif event == 'GUILD_BAN_ADD':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_ban', guild_id, await User.from_api_res(data))

        elif event == 'GUILD_BAN_REMOVE':
            guild_id = data['guild_id']
            return await self.http.get_client().raise_event('on_ban_remove', guild_id, await User.from_api_res(data))

        elif event == 'TYPING_START':
            await self.http.get_client().raise_event('on_typing_start', data['user_id'], data['channel_id'], data['timestamp'])

        elif event == 'MESSAGE_CREATE':
            message = await ChannelMessage.from_api_res(data)
            await self.http.get_client().raise_event('on_message', message)

        elif event == 'MESSAGE_UPDATE':
            message = await ChannelMessage.from_api_res(data)
            await self.http.get_client().raise_event('on_message_create', message)

        elif event == 'MESSAGE_DELETE':
            await self.http.get_client().raise_event('on_message_delete', data['id'], data['channel_id'])

        elif event == 'MESSAGE_DELETE_BULK':
            await self.http.get_client().raise_event('on_message_delete_bulk', data['ids'], data['channel_id'])

        elif event == 'MESSAGE_REACTION_ADD':
            emoji = await Emoji.from_api_res(data['emoji'])
            await self.http.get_client().raise_event(
                'on_message_reaction_add', data['user_id'], data['channel_id'], data['message_id'], emoji)

        elif event == 'MESSAGE_REACTION_REMOVE':
            emoji = await Emoji.from_api_res(data['emoji'])
            await self.http.get_client().raise_event(
                'on_message_reaction_remove', data['user_id'], data['channel_id'], data['message_id'], emoji)

        elif event == 'MESSAGE_REACTION_REMOVE_ALL':
            await self.http.get_client().raise_event(
                'on_message_reaction_remove_all', data['channel_id'], data['message_id'])

        else:
            logger.critical(
                f'Unhandled event type {event}, data: {data}')
